import inspect
import json
from abc import ABC, abstractmethod
from typing import (Any, Callable, Coroutine, List, Optional, Set, Type,
                    TypeVar, re)

import openai
from pydantic import BaseModel  # pylint: disable=no-name-in-module
from pydantic import create_model  # pylint: disable=no-name-in-module
from typing_extensions import ParamSpec

from ..utils import Caser

F = TypeVar("F", bound="OpenAIFunction")
T = TypeVar("T")
P = ParamSpec("P")
c = Caser()

class FunctionCall(BaseModel):
	name: str
	data: Any

class OpenAIFunction(BaseModel, ABC):
	"""
	OpenAI Function
	---------------

	OpenAI Function orchestration, the `json_schema` needed for the OpenAI API is automatically generated, all the subclasses of this class are automatically registered and can be used in the `function_call` coroutine
	"""
	class Metadata:
		subclasses: Set[Type[F]] = set()

	@classmethod
	def __init_subclass__(cls, **kwargs: Any):
		super().__init_subclass__(**kwargs)
		_schema = cls.schema()
		if cls.__doc__ is None:
			cls.__doc__ = ""
		cls.openaischema = {
			"name": cls.__name__,
			"description": cls.__doc__,
			"parameters": {
				"type": "object",
				"properties": {
					k: v for k, v in _schema["properties"].items() if k != "self"
				},
				"required": _schema.get("required", []),
			},
		}
		cls.Metadata.subclasses.add(cls)

	@classmethod
	def from_function(cls, func: Callable[P, Coroutine[Any,Any,Any]]) -> Type["OpenAIFunction"]:
		"""
		Creates an OpenAIFunction from a coroutine and infers it's schema from the function signature
		"""
		
		kwargs = inspect.signature(func).parameters
		
		klass = create_model(
		   c(func.__name__,"pascal"),
			__base__=cls,
			**{
				k: (v.annotation, ...) for k, v in kwargs.items() if k != "self"
			},
		)

		klass.__doc__ = func.__doc__
		
		
		async def run(self, **kwargs: Any) -> Any:
			return await func(self, **kwargs)
		
		setattr(klass, "run", run)
		klass.__init_subclass__()
		
		return klass
	
	async def __call__(self, **kwargs: Any) -> FunctionCall:
		response = await self.run(**kwargs)
		return FunctionCall(name=self.__class__.__name__, data=response)

	async def run(self, **kwargs: Any) -> Any:
		"""Main function to be implemented by subclasses"""
		...

async def parse_openai_function(
	response: dict[str, Any],
	functions: List[Type[F]] = OpenAIFunction.Metadata.subclasses, 
	**kwargs: Any,
) -> FunctionCall:
	"""Parses the response from the OpenAI API and returns a FunctionCall object, meant to have an uniform interface"""
	choice = response["choices"][0]["message"]
	if "function_call" in choice:
		function_call_ = choice["function_call"]
		name = function_call_["name"]
		arguments = function_call_["arguments"]
		print(name, arguments)
		for i in functions:
			if i.__name__ == name:
				result = await i(**json.loads(arguments))(**kwargs)
				break
		else:
			raise ValueError(f"Function {name} not found")
		return result
	return FunctionCall(name="chat", data=choice["content"])


async def function_call(
	text: str,
	model: str = "gpt-3.5-turbo-16k-0613",
	functions: List[Type[F]] = OpenAIFunction.Metadata.subclasses,
	**kwargs: Any,
) -> FunctionCall:
	"""Given the user input, it will infer the appropiate function to call and it's `json_schema`, then it will call the function and return a FunctionCall object"""
	messages = [
		{"role": "user", "content": text},
		{"role": "system", "content": "You are a function Orchestrator"},
	]
	response = await openai.ChatCompletion.acreate(  # type: ignore
		model=model,
		messages=messages,
		functions=[i.openaischema for i in functions],
	) 
	return await parse_openai_function(response, functions=functions, **kwargs) # type: ignore


async def chat_completion(text: str, context: Optional[str] = None) -> str:
	"""

	Chat Completion
	---------------

	Args:

		text (str): The text to generate from

		context (Optional[str], optional): The context to generate from. Defaults to None.

	Returns:

		str: The generated text

	Simple chat completion outside agent context
	"""
	
	if context is not None:
		messages = [
			{"role": "user", "content": text},
			{"role": "system", "content": context},
		]
	else:
		messages = [{"role": "user", "content": text}]
	response = await openai.ChatCompletion.acreate(
		model="gpt-3.5-turbo-16k-0613", messages=messages
	)
	return response["choices"][0]["message"]["content"]


async def instruction(text: str, temperature: float = 0.2, max_tokens: int = 1024)->str:
	"""

	Instruction Completion
	---------------

	Args:

		text (str): The text to generate from

	Returns:

		str: The generated text

	Simple instruction completion outside agent context
	"""
	response = await openai.Completion.acreate( # type: ignore
			model="gpt-3.5-turbo-instruct",
			prompt=text,
			temperature=temperature,
			max_tokens=max_tokens,
			stream=False
		)
	return response.choices[0].text # type: ignore



import functools
from typing import Callable, Coroutine, Dict, List, Literal, TypeAlias, TypeVar

import openai  # pylint-disable=E0401
from aiofauna import *  # type: ignore
from jinja2 import Template
from pydantic import BaseModel, Field  # pylint: disable=no-name-in-module
from typing_extensions import ParamSpec

T = TypeVar("T")
P = ParamSpec("P")

Role: TypeAlias = Literal["assistant", "user", "system", "function"]
Model: TypeAlias = Literal["gpt-4-0613", "gpt-3.5-turbo-16k-0613"]
Size: TypeAlias = Literal["256x256", "512x512", "1024x1024"]
Format: TypeAlias = Literal["url", "base64"]


class Message(Document):
	"""
	A message in a chat

	Args:
		role (Role): The role of the message (assistant, user, system, function)
		content (str): The content of the message
	"""
	role: Role = Field(..., description="The role of the message")
	content: str = Field(..., description="The content of the message")


class ChatCompletionRequest(Document):
	"""

	Chat completion request

	Args:

		model (Model): The model to use for the completion (gpt-4-0613, gpt-3.5-turbo-16k-0613)
		messages (List[Message]): The messages to use for the completion
		temperature (float): The temperature of the completion
		max_tokens (int): The maximum number of tokens to generate
		stream (bool): Whether to stream the completion or not
	"""
	model: Model = Field(..., description="The model to use for the completion")
	messages: List[Message] = Field(
		..., description="The messages to use for the completion"
	)
	temperature: float = Field(
		default=0.5, description="The temperature of the completion"
	)
	max_tokens: int = Field(
		1024, description="The maximum number of tokens to generate"
	)
	stream: bool = Field(False, description="Whether to stream the completion or not")


class ChatCompletionUssage(Document):
	"""

	Chat completion usage

	Args:

		prompt_tokens (int): The number of tokens in the prompt

		completion_tokens (int): The number of tokens in the completion

		total_tokens (int): The total number of tokens

	"""
		
	prompt_tokens: int = Field(..., description="The number of tokens in the prompt")
	completion_tokens: int = Field(
		..., description="The number of tokens in the completion"
	)
	total_tokens: int = Field(..., description="The total number of tokens")


class ChatCompletionChoice(Document):
	"""

	Chat completion choice

	Args:

		index (int): The index of the choice

		message (Message): The message of the choice

		finish_reason (str): The reason the choice was finished

	""" 
	
	index: int = Field(..., description="The index of the choice")
	message: Message = Field(..., description="The message of the choice")
	finish_reason: str = Field(..., description="The reason the choice was finished")


class ChatCompletionResponse(Document):
	"""

	Chat completion response

	Args:

		id (str): The id of the completion

		object (str): The object of the completion

		created (int): The creation time of the completion

		model (Model): The model used for the completion

		choices (List[ChatCompletionChoice]): The choices of the completion

		usage (ChatCompletionUssage): The usage of the completion

		stream (bool): Whether the completion was streamed or not

	"""

	id: str = Field(..., description="The id of the completion")
	object: str = Field(..., description="The object of the completion")
	created: int = Field(..., description="The creation time of the completion")
	model: Model = Field(..., description="The model used for the completion")
	choices: List[ChatCompletionChoice] = Field(
		..., description="The choices of the completion"
	)
	usage: ChatCompletionUssage = Field(..., description="The usage of the completion")
	stream: bool = Field(..., description="Whether the completion was streamed or not")


class VectorResponse(Document):
	"""

	Vector response

	Args:

		id (str): The id of the vector

		object (str): The object of the vector

		created (int): The creation time of the vector

		model (Model): The model used for the vector

		data (List[float]): The data of the vector

	"""  
	
	text: str = Field(..., description="The text of the completion")
	score: float = Field(..., description="The score of the completion")


class CreateImageResponse(Document):
	"""

	Create image response

	Args:

		id (str): The id of the image

		object (str): The object of the image

		created (float): The creation time of the image

		data (List[Dict[Format, str]]): The data of the image

	"""
	
	created: float = Field(...)
	data: List[Dict[Format, str]] = Field(...)


class CreateImageRequest(Document):
	"""

	Create image request

	Args:

		prompt (str): The prompt of the image

		n (int): The number of images to generate

		size (Size): The size of the image

		response_format (Format): The format of the response

	"""   
	
	prompt: str = Field(...)
	n: int = Field(default=1)
	size: Size = Field(default="1024x1024")
	response_format: Format = Field(default="url")


class LanguageModel(BaseModel):
	"""

	ChatGPT: A GPT-3.5+ chatbot

	This class represents ChatGPT Large Language Model behavior and is a wrapper around the OpenAI Chat API.

	Args:

		model (Model): The model to use for the completion (gpt-4-0613, gpt-3.5-turbo-16k-0613)

		temperature (float): The temperature of the completion

		max_tokens (int): The maximum number of tokens to generate

	"""  
	
	
	model: Model = Field(default="gpt-4-0613")
	temperature: float = Field(default=0.2)
	max_tokens: int = Field(default=1024)

	async def chat(
		self, text:str, context:Optional[str]=None
	) -> str:
		"""

		Chat with the bot with the option to pass a context

		Args:

			text (str): The text to send to the bot

			context (Optional[str], optional): The context of the bot. Defaults to None.

		Returns:

			str: The response of the bot

		"""

		if context is not None:
			messages = [
				Message(role="user", content=text).dict(),
				Message(role="system", content=context).dict(),
			]
		else:
			messages = [Message(role="user", content=text).dict()]
	   
		response = await openai.ChatCompletion.acreate(  # type: ignore
			model=self.model,
			messages=messages,
			temperature=self.temperature,
			max_tokens=self.max_tokens,
			stream=False
		)
		
		return response.choices[0].message.content  # type: ignore

	def __call__(self, func: Callable[P,Coroutine[T,Any,Any]]) -> Callable[P,Coroutine[T,Any,Any]]:
		"""

		Call the bot with a function that features a jinja2 docstring as a template for the system message, and the user message as the first argument, prompt engineering is done automatically

		Args:

			func (Callable[P,Coroutine[T,Any,Any]]): The function to call

		Returns:

			Callable[P,Coroutine[T,Any,Any]]: The wrapped function

		"""
		if func.__doc__ is None:
			raise ValueError("Function must have a docstring")
		_template = func.__doc__	
		@handle_errors
		@functools.wraps(func)
		async def wrapper(*args:P.args, **kwargs:P.kwargs) -> str:
			template = Template(_template)
			_result = await func(*args, **kwargs)
			if isinstance(_result, dict):
				template_kwargs= _result # type: ignore
			elif isinstance(_result, BaseModel):
				template_kwargs = _result.dict()
			elif isinstance(_result, str):
				template_kwargs = {"text": _result}
			else:
				raise ValueError("Function must return a dict, a BaseModel or a str")
			assert isinstance(template_kwargs["text"], str)
			context = template.render(**template_kwargs)
			return await self.chat(text=template_kwargs["text"], context=context)
		return wrapper
	
	async def assist(self, text:str)->str:
		"""
		Assistance with `gpt-3.5-turbo-instruct` model. Similar to `text-davinci` but with more focus on instructions and state of the GPT3.5 model.

		Args:

			text (str): The text to send to the bot

		Returns:

			str: The response of the bot

					"""
		response = await openai.Completion.acreate( # type: ignore
			model="gpt-3.5-turbo-instruct",
			prompt=text,
			temperature=self.temperature,
			max_tokens=self.max_tokens,
			stream=False
		)
		return response.choices[0].text # type: ignore
	

	async def stream(self, text:str, context:Optional[str]=None)->AsyncGenerator[str, None]:
		"""
		The chat between the bot and the user is streamed through an async generator, suitable for websockets and server-sent events

		Args:

			text (str): The text to send to the bot

			context (Optional[str], optional): The context of the bot. Defaults to None.
	
		Yields:

			AsyncGenerator[str, None]: The response of the bot

		"""

		if context is not None:
			messages = [
				Message(role="user", content=text).dict(),
				Message(role="system", content=context).dict(),
			]
		else:
			messages = [Message(role="user", content=text).dict()]
	   
		response = await openai.ChatCompletion.acreate(  # type: ignore
			model=self.model,
			messages=messages,
			temperature=self.temperature,
			max_tokens=self.max_tokens,
			stream=True
		)

		async for message in response:
			data = message["choices"][0]["delta"].get("content")
			if data is not None:
				yield data
			else:
				break


from dataclasses import dataclass, field

from .memory import *
from .tools import *


@dataclass(
	frozen=True,
	order=True,
	unsafe_hash=True
)
class Agent(ABC):
	
	
	llm:LanguageModel = field(default=LanguageModel())
	memory:Memory = field(default=Memory())
	namespace:str = field(default="main")
	n:int = field(default=5)
	

	@abstractmethod
	async def chain(self, text:str)->str:
		"""
		Chain a text
		"""
		...

class Swagchain(Agent):
	"""
	Swagchain: A retrieval augmented generation agent
	----------------------

	This class is a wrapper around the OpenAI Chat API and Pinecone Embeddings API
	"""

	async def chain(self, text:str)->str:
		"""
		Retrieval Augmented Generation
		----------------------
		
		Args:
		
			text (str): The text to retrieve from

		Returns:

			str: The generated text
		
		The agent will find for the KNN of the text into his memory namespace and generate from them
		a response
		"""

		knn = await self.memory.search(text=text, namespace=self.namespace, top_k=self.n)
		if len(knn) == 0:
			response = await self.llm.chat(text=text)
		else:
			context = "Suggestions:\n" + "\n".join(knn)
			response = await self.llm.chat(text=text, context=context)
		await self.memory.save(texts=[text, response], namespace=self.namespace)
		return response
	

	async def stream_chain(self, text:str)->AsyncGenerator[str, None]:
		"""
		Retrieval Augmented Generation (stream)
		----------------------
		
		Args:
		
			text (str): The text to retrieve from

		Returns:

			AsyncGenerator[str, None]: The generated text
		
		The agent will find for the KNN of the text into his memory namespace and generate from them
		a response
		"""
		full_response = ""
		knn = await self.memory.search(text=text, namespace=self.namespace, top_k=self.n)
		if len(knn) == 0:
			async for response in self.llm.stream(text=text):
				full_response += response
				yield response
		else:
			context = "Suggestions:\n" + "\n".join(knn)
			async for response in self.llm.stream(text=text, context=context):
				full_response += response
				yield response
		await self.memory.save(texts=[text, full_response], namespace=self.namespace)


	async def assist(self,text:str)->str:
		"""
		Assist with GPT-3.5 instructions
		----------------------

		Args:

			text (str): The text to assist with

		Returns:

			str: The generated text

		The agent will call the GPT-3.5 instruction completion API
		"""

		return await self.llm.assist(text=text)


import asyncio
from typing import AsyncGenerator, List

from aiofauna import handle_errors, process_time, setup_logging  # type: ignore
from aiohttp import ClientSession, TCPConnector
from bs4 import BeautifulSoup  # pylint: disable=import-error

from ..llm import *
from ..memory import *

logger = setup_logging(__name__)

HEADERS = {
	"User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) "
	"AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.114 Safari/537.36"
}
BAD_EXT = (
	"png",
	"jpg",
	"jpeg",
	"gif",
	"pdf",
	"doc",
	"docx",
	"ppt",
	"pptx",
	"xls",
	"xlsx",
	"zip",
	"rar",
	"gz",
	"7z",
	"exe",
	"mp3",
	"mp4",
	"avi",
	"mkv",
	"mov",
	"wmv",
	"flv",
	"swf",
)

connector = TCPConnector(limit=1000)

logger = setup_logging(__name__)



@process_time
@handle_errors
async def sitemap_urls(url: str, session: ClientSession) -> List[str]:
	"""
	Fetches all urls from a sitemap
	"""
	urls: List[str] = []
	if not url.endswith("xml"):
		url = f"{url.rstrip('/')}/sitemap.xml"
	async with session.get(url) as response:
		text = await response.text()
		soup = BeautifulSoup(text, features="xml")
		for loc in soup.findAll("loc"):
			if loc.text.endswith(BAD_EXT):
				continue
			urls.append(loc.text)
			logger.info("Found url: %s", loc.text)
		for nested_sitemap in soup.findAll("sitemap"):
			urls.extend(await sitemap_urls(nested_sitemap.loc.text, session))
	return urls


@process_time
@handle_errors
async def fetch_website(url: str, session: ClientSession, max_size: int = 40960) -> str:
	"""
	Fetches a single website
	"""
	async with session.get(url) as response:
		html = await response.text()
		soup = BeautifulSoup(html, features="lxml")
		tag = soup.find("div", class_="content")
		assert tag is not None
		data = tag.get_text(strip=True)
		logger.info("Fetched %s", url)
		logger.info("Length: %s", len(data))
		return data[:max_size]


async def sitemap_pipeline(
	url: str,
	namespace: str,
	session: ClientSession,
	memory: Memory = Memory(),
	chunk_size: int = 64,
)->AsyncGenerator[float, None]:
	"""
	Fetches all urls from a sitemap
	"""
	urls: List[str] = await sitemap_urls(url, session)
	length: int = len(urls)
	inserted: int = 0
	while urls:
		try:
			chunk = urls[:chunk_size]
			urls = urls[chunk_size:]
			contents = await asyncio.gather(
				*[fetch_website(url, session) for url in chunk]
			)
			inserted += await memory.save(texts=contents, namespace=namespace)
			progress = inserted / length
			logger.info("Progress: %s", progress)
			yield progress
			if progress >= 1:
				logger.info("Succesfully inserted %s documents", inserted)
				break
		except Exception as exc: # pylint: disable=broad-except
			logger.error(exc)
			continue
			

@process_time
@handle_errors
async def scrap_urls(url: str, session: ClientSession) -> List[str]:
	"""
	Find all children of a website
	"""
	async with session.get(url + "/index.html") as response:
		html: str = await response.text()
		soup = BeautifulSoup(html, features="lxml")
		results = soup.find_all("a", href=True)
		logger.info("Found %s results", len(results))
		responses: List[str] = []
		for result in results:
			if result["href"].startswith("http"):
				responses.append(result["href"])
			else:
				responses.append(f"{url.rstrip('/')}/{result['href'].lstrip('/')}")
		return responses


@process_time
@handle_errors
async def scrape_urls_recursive(url: str, session: ClientSession) -> List[str]:
	urls = await scrap_urls(url, session)
	"""
	Find all children of a website recursively
	"""
	for url in urls:
		urls.extend(await scrape_urls_recursive(url, session))
	return urls


async def scrape_pipeline(
	url: str,
	namespace: str,
	session: ClientSession,
	memory: Memory = Memory(),
	chunk_size: int = 64,
):
	"""
	Scrapes a website
	"""
	urls = await scrap_urls(url, session)
	logger.info("Found %s urls", len(urls))
	logger.info("Fetching all children")
	length = len(urls)
	inserted = 0
	while urls:
		chunk = urls[:chunk_size]
		urls = urls[chunk_size:]
		try:
			contents = await asyncio.gather(
				*[fetch_website(url, session) for url in chunk]
			)
			inserted += await memory.save(
				texts=contents, namespace=namespace
			)
			progress = inserted / length
			logger.info("Progress: %s", progress)
			yield progress
			if progress >= 1:
				logger.info("Succesfully inserted %s documents", inserted)
				break
		except Exception as exc: # pylint: disable=broad-except
			logger.error(exc)
			continue


async def website_loader(url: str, namespace: str) -> AsyncGenerator[float, None]:
	"""Reads a website and returns a list of strings"""
	async with ClientSession(connector=connector, headers=HEADERS) as session:
		exts = ["sitemap.xml", "index.html", ""]
		for e in exts:
			response = await session.get(url + "/" + e)
			if response.status == 200:
				if e == "sitemap.xml":
					async for progress in sitemap_pipeline(url, namespace, session):
						yield progress
				else:
					async for progress in scrape_pipeline(url, namespace, session):
						yield progress
				break
		else:
			logger.error("Could not find a sitemap.xml or index.html file")
			raise Exception("Could not find a sitemap.xml or index.html file")
